/**************************************************************************
 
 	asteroid.c 
 
  	Project: RockBlast an asteroid game. 
 
	Copyright (c) Febuary 2001 Garett Spencley
 
   	This program is free software; you can redistribute it and/or modify
   	it under the terms of the GNU General Public License as published by
   	the Free Software Foundation; either version 2, or (at your option)
   	any later version.

   	This program is distributed in the hope that it will be useful,
   	but WITHOUT ANY WARRANTY; without even the implied warranty of
   	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   	GNU General Public License for more details.

   	You should have received a copy of the GNU General Public License
   	along with this program; if not, write to the Free Software Foundation,
   	Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 

****************************************************************************/

#include <stdlib.h>
#include <string.h>

#include "asteroid.h"
#include "level.h"
#include "image.h"
#include "score.h"
#include "audio.h"
#include "lives.h"

#define X_OFFSET 20
#define Y_OFFSET 15

unsigned int SPECIAL[] = {  /* The "special" asteroids */
	ONE_UP, BONUS1, BONUS2, TEMP_BLOCKER, KILLER 
};

SDL_Surface *asteroid_image[NUM_TYPES];
int ASTEROIDS_LEFT;

bool is_special(unsigned int type);
char *get_asteroid_image(unsigned int type);
unsigned int get_rand_special(void);

/*
 * free_images(n) - free all the asteriod images up to 'n'
 */

inline static void free_images(int n)
{
	int i;

	for(i = 0; i < n; i++) 
		SDL_FreeSurface(asteroid_image[i]);
}

/*
 * init_asteroids() - load data for asteroids
 */

bool init_asteroids()
{
	char *path, *file;
	int i;

	for(i = 0; i < NUM_TYPES; i++) {

		file = get_asteroid_image(i);
		if(!file) {
			fprintf(stderr, "out of memory\n");
			free_images(i);
			return false;
		}

		path = data_file(file);
		if(!path) {
			fprintf(stderr, "out of memory\n");
			free(file);
			free_images(i);
			return false;
		}
		free(file);

		asteroid_image[i] = load_image(path, true);
		if(!asteroid_image[i]) {
			fprintf(stderr, "Unable to load image %s: %s\n", path, SDL_GetError());
			free_images(i);
			return false;
		}

		free(path);
	}

	return true;
}

/* 
 * get_asteroid_image() - return the image file for 'type'
 */

char *get_asteroid_image(unsigned int type)
{
	switch(type) {
		case NORMAL: 		return strdup("asteroid.png");
		case BLOCKER: 		return strdup("block_asteroid.png");
		case TEMP_BLOCKER: 	return strdup("tempblock.png");
		case ONE_UP: 		return strdup("oneup.png");
		case KILLER: 		return strdup("killer.png");
		case BONUS1: 		return strdup("bonus1.png");
		case BONUS2: 		return strdup("bonus2.png");
		default:			return NULL;
	}

	return NULL;
}

/*
 * create_asteroids() - create all the asteroids
 */

bool create_asteroids(asteroid_t *asteroids, int num)
{
	int i;
	int alive, type;
	int yval, xval;

	xval = X_OFFSET;
	yval = Y_OFFSET;

	ASTEROIDS_LEFT = -1;

	for(i = 0; i < num; i++) {

		/* Determine type of asteroid. Blocker or normal */

		if(difficulty_level == 0)
			type = NORMAL;
		else if(difficulty_level < 3)
			type = rand() % (5 - difficulty_level);	
		else 
			type = rand() % 2;

		if(type != BLOCKER) type = NORMAL;

		asteroids[i].obj.image = asteroid_image[type];
		asteroids[i].type = type;
		asteroids[i].hits = 0;

		/* Determine position of asteroid */

		if(i != 0) {
				
			if(i % 20 == 0) {
				yval = asteroids[i-1].obj.y + asteroids[i-1].obj.image->h + Y_OFFSET;
				xval = X_OFFSET;
			} else 
				xval = asteroids[i-1].obj.x + asteroids[i-1].obj.image->w + 10;
		}

		asteroids[i].obj.x = xval;
		asteroids[i].obj.y = yval;

		/* Determine wether the asteroid should be alive or not */

		alive = rand() % 2;

		if(alive) {
			++ASTEROIDS_LEFT;
			asteroids[i].alive = true;
		} else
			asteroids[i].alive = false; 
	}

	return true;
}

/* 
 * free_asteroids() - destroy asteroid data
 */

void free_asteroids(asteroid_t *asteroids, int num)
{
	free_images(NUM_TYPES);
}

/*
 * draw_asteroids() - draw the asteroids
 */

inline void draw_asteroids(asteroid_t *asteroids, int num)
{
	int i;
	time_t cur;

	for(i = 0; i < num; i++) {
		
		if(is_special(asteroids[i].type)) {

			cur = time(NULL);

			if((cur - asteroids[i].ts) >= SPECIAL_TIME_LIMIT)
				asteroids[i].alive = false;
		}
		
		if(asteroids[i].alive)
			draw_object(&asteroids[i].obj); 
	}
}

/*
 * erase_asteroids() - erase the asteroids
 */

void erase_asteroids(asteroid_t *asteroids, int num)
{
	int i;

	for(i = 0; i < num; i++)
		if(asteroids[i].alive)
			erase_object(&asteroids[i].obj);
}

/* 
 * create_special() - randomly decide wether to create a special 
 * asteroid or not 
 */

inline bool create_special()
{
	int create;

	/* to create or not to create? */
	create = rand() % 300;

	if(create == 0)
		return true;

	return false;
}

/*
 * special_asteroid() - randomly create special asteroids
 */

void special_asteroid(asteroid_t *asteroids, int num)
{
	int i, start, type;

	/* Try to randomize where the asteroid will be */

	do {
		start = (rand() % num) + 1;
	} while(start == num);

	/* Look for a free slot */
	for(i = start; i < num; i++) {
		if(!asteroids[i].alive)
			break;
	}
	
	if(i == num) 
		return;

	/* Pick a random type. Leave killer asteroids
	 * out until level 3 */

	type = get_rand_special();

	asteroids[i].obj.image = asteroid_image[type];
	asteroids[i].type = type;

	asteroids[i].ts = time(NULL);

	asteroids[i].alive = true;
}

/* 
 * is_special() - returns true if 'type' is a special asteroid
 */

bool is_special(unsigned int type)
{
	int i;

	for(i = 0; i < ARRAYSIZE(SPECIAL); i++) 
		if(type == SPECIAL[i]) 
			return true;	
	
	return false;
}

/* 
 * get_rand_special() - returns a random special asteroid
 */

unsigned int get_rand_special()
{
	unsigned int type;

	if(difficulty_level <= 2) {
		do {
			type = SPECIAL[rand() % ARRAYSIZE(SPECIAL)];
		} while (type == KILLER);
	} else 
		type = SPECIAL[rand() % ARRAYSIZE(SPECIAL)];

	return type;
}

/* 
 * asteroid_score() - add or subtract the appropriate score
 * for hitting an asteroid of 'type'
 */

void asteroid_score(unsigned int type)
{
	switch(type) {

	case NORMAL:
	case BLOCKER:
		add_score(5);
		break;

	case ONE_UP:
		break;

	case KILLER:
		sub_score(10);
		break;

	case BONUS1:
		add_score(50);
		break;
	
	case BONUS2:
		add_score(100);
		break;
	}
}

/*
 * ball_meets_asteroid() - perform appropriate action for
 * when the ball hits an asteroid of 'type'
 */

void ball_meets_asteroid(ball_t *ball, asteroid_t *asteroid)
{
	switch(asteroid->type) {

	case TEMP_BLOCKER:
		play_sample(BUMP);
		break;

	case BLOCKER:

		if(++asteroid->hits == 2) {

			asteroid->alive = false;
			--ASTEROIDS_LEFT;
				
			play_sample(BLOWUP);
	
		} else
			play_sample(BUMP);

		break;

	case ONE_UP:
		add_life(ball);
		asteroid->alive = false;

		play_sample(ONE_UP);
		break;

	case KILLER:
		asteroid->alive = false;

		kill_ball(ball);

		play_sample(DIE);
		break;

	case BONUS1:
	case BONUS2:
		asteroid->alive = false;
		play_sample(BONUS);
		break;

	case NORMAL:
		--ASTEROIDS_LEFT;
		asteroid->alive = false;
			
		play_sample(BLOWUP);

		break;
	}

	asteroid_score(asteroid->type);
}

