/**************************************************************************
 
 	ball.c 
 
  	Project: RockBlast an asteroid game. 
 
	Copyright (c) Febuary 2001 Garett Spencley
 
   	This program is free software; you can redistribute it and/or modify
   	it under the terms of the GNU General Public License as published by
   	the Free Software Foundation; either version 2, or (at your option)
   	any later version.

   	This program is distributed in the hope that it will be useful,
   	but WITHOUT ANY WARRANTY; without even the implied warranty of
   	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   	GNU General Public License for more details.

   	You should have received a copy of the GNU General Public License
   	along with this program; if not, write to the Free Software Foundation,
   	Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 

****************************************************************************/

#include <math.h>
#include <stdio.h>

#include "ball.h"
#include "asteroid.h"
#include "paddle.h"
#include "level.h"
#include "lives.h"
#include "audio.h"
#include "score.h"

#define BALL_SPEED 6

#define MIN_DY_BALL 2
#define MIN_DX_BALL 2

#define MAX_X_VEL 14
#define MAX_Y_VEL 14

typedef enum { 
	NO_HIT, TOP_LEFT, TOP_RIGHT, 
	BOTTOM_LEFT, BOTTOM_RIGHT 
} region;

region hit_asteroid(ball_t *, asteroid_t *);
bool hit_paddle(ball_t *);
void check_asteroid_collision(ball_t *, asteroid_t *); 

/*
 * move_ball() - check for collisions and move the ball
 */

void move_ball(ball_t *ball)
{
	float vx, vy, vs;
	float alpha, gamma, beta;
	float paddle_size;

	/* Check for collision with side walls */

	if((ball->obj.x + ball->obj.image->w) >= (screen->w - ball->obj.image->w)) 	
		ball->dx = -(abs(ball->dx));
	else if(ball->obj.x <= 0)
		ball->dx = abs(ball->dx);	

	/* Check for collision with top wall or the bottom */
	
	if((ball->obj.y + ball->obj.image->h) >= screen->h) {
		
		/* ball fall down go boom */
		kill_ball(ball);	
		play_sample(DIE);

	} else if((ball->obj.y - ball->obj.image->h) <= 0)
		ball->dy = abs(ball->dy);

	/* Check for collision with paddle */
		
	if(hit_paddle(ball)) {
			
		/* Reminder: This code was borrowed from xboing 2.0 */
	
		/* speed vector of the ball */
		vx = (float)ball->dx;
		vy = (float)ball->dy;

		/* speed intensity of the ball */
		vs = (float)sqrt(vx * vx + vy * vy);
			
		alpha = (float)atan(vx / -vy);

		paddle_size = (float)paddle.obj.image->w;
		vx = (float)ball->obj.x - paddle.obj.x;
		vy = (float)(paddle_size / 1.0);

		beta = (float)atan(vx / vy);
		gamma = (float)(1.0 * beta - alpha);
					
		vx = (float)(vs * sin(gamma));
		vy = (float)(-vs * cos(gamma));
		   
		/* Take into account horizontal speed of the paddle */
		vx += (float)(paddle.dx / 10.0);

		if(vx > 0.0) 
			ball->dx = (int)(vx + 0.5);
		else
			ball->dx = (int)(vx - 0.5);

		if(vy < 0.0)
			ball->dy = (int)(vy - 0.5);
		else
			ball->dy = -MIN_DY_BALL;

		if(ball->dy > -MIN_DY_BALL)
			ball->dy = -MIN_DY_BALL;
	}

	vx = (float)ball->dx;
   	vy = (float)ball->dy;
   	vs = (float)sqrt(vx * vx + vy * vy);

   	alpha = (float)sqrt(MAX_X_VEL * MAX_X_VEL + MAX_Y_VEL * MAX_Y_VEL);
		
    alpha /= 9.0; /* number of speed level */
    alpha *= (float)speed_level;
		
	if(vs == 0.0) 
		vs = 1.0;
		
    beta = alpha / vs;

    vx *= beta;
   	vy *= beta;

    if(vx > 0.0)
       	ball->dx = (int)(vx + 0.5);
    else
       	ball->dx = (int)(vx - 0.5);

    if(vy > 0.0)
      	ball->dy = (int)(vy + 0.5);
    else
       	ball->dy = (int)(vy - 0.5);

   	if(ball->dy == 0) 
		ball->dy = MIN_DY_BALL;

   	if(ball->dx == 0) 
		ball->dx = MIN_DX_BALL;
	
	/* Check for asteroid collision */

	check_asteroid_collision(ball, asteroids);
	
	/* Actually move the ball. */

	if(ball->moving) {
		ball->obj.x += ball->dx * BALL_SPEED;
		ball->obj.y += ball->dy * BALL_SPEED;
	}

	/* FIXME: Important safety checks but could be done better (for speed)
	 * when dx and dy are computed. */
	
	if((ball->obj.x + ball->obj.image->w) >= screen->w)
		ball->obj.x = screen->w - ball->obj.image->w;
	else if(ball->obj.x <= 0)
		ball->obj.x = 0;
	
	if(ball->obj.y <= 0) 
		ball->obj.y = 0;
	else if(ball->obj.y + ball->obj.image->h >= screen->h)
		ball->obj.y = screen->h - ball->obj.image->h;
}

/* 
 * check_asteroid_collision() - check for collision with asteroid
 * and change direction of ball appropriately 
 */

void check_asteroid_collision(ball_t *ball, asteroid_t *asteroids)
{
	int i;
	region part;

	for(i = 0; i < NUM_ASTEROIDS; i++) {

		part = hit_asteroid(ball, &asteroids[i]);
		
		if(part != NO_HIT) {

			ball_meets_asteroid(ball, &asteroids[i]);

			/* Bounce the ball off in an appropriate direction */
			
			switch(part) {
			
			case BOTTOM_LEFT:		
				ball->dy = abs(ball->dy);
				ball->dx = -(abs(ball->dx));
				break;

			case BOTTOM_RIGHT:
				ball->dy = abs(ball->dy);
				ball->dx = abs(ball->dx);
				break;
				
			case TOP_LEFT:
				ball->dy = -(abs(ball->dy));
				ball->dx = -(abs(ball->dx));
				break;

			case TOP_RIGHT:
				ball->dy = -(abs(ball->dy));
				ball->dx = abs(ball->dx);
				break;

			default:
			case NO_HIT:
				break;
			}
		}
	}
}

/*
 * draw_ball() - draw the ball
 */

inline void draw_ball(ball_t *ball)
{
	draw_object(&ball->obj);
}

/*
 * erase_ball() - erase the ball
 */

inline void erase_ball(ball_t *ball)
{
	erase_object(&ball->obj);
}

/*
 * hit_asteroid() - check for collision with asteroid
 * returning region of asteroid that was hit
 */

region hit_asteroid(ball_t *ball, asteroid_t *asteroid)
{
	object *tmp;
	int ball_x, ball_y = 0;
	int left_half, bottom_half = 0;
	
	if(!asteroid->alive)
		return NO_HIT;
	
	tmp = &asteroid->obj;
	
	/* Did it hit? */
	
	if((ball->obj.x + ball->obj.image->w  >= tmp->x && 
		ball->obj.x <= (tmp->x + tmp->image->w)) &&
		(ball->obj.y + ball->obj.image->w) >= tmp->y && 
		ball->obj.y <= (tmp->y + tmp->image->h)) {

		/* Sure did :O) But where? */
	
		ball_x = ball->obj.x + ball->obj.image->w;
		ball_y = ball->obj.y; // + ball->obj.image->h;
		
		left_half = (tmp->x + tmp->image->w) - (tmp->image->w / 2);
		bottom_half = (tmp->y + tmp->image->h) - (tmp->image->h / 2);
		
		if(ball_x < left_half) {
			if(ball_y < bottom_half)
				return TOP_LEFT;
			else
				return BOTTOM_LEFT;
		} else {
			if(ball_y < bottom_half)
				return TOP_RIGHT;
			else
				return BOTTOM_RIGHT;
		}
	}
	
	return NO_HIT;	
}

/*
 * hit_paddle() - returns true if ball hit paddle
 */

bool hit_paddle(ball_t *ball)
{
	int ball_x, ball_y;

	ball_x = ball->obj.x + ball->obj.image->w;
	ball_y = ball->obj.y + ball->obj.image->h;
	
	if(ball_x >= paddle.obj.x && 
		ball->obj.x <= (paddle.obj.x + paddle.obj.image->w) && // + ball->obj.image->w) &&
		(ball_y >= paddle.obj.y && ball_y <= paddle.obj.y + paddle.obj.image->h))
			return true;

	return false;
}

/*
 * kill_ball() - kill the ball 
 */

void kill_ball(ball_t *ball)
{
	take_life(ball);

	ball->moving = false;

	ball->obj.x = paddle.obj.x + (paddle.obj.image->w / 2);
	ball->obj.y = (paddle.obj.y - ball->obj.image->h);
		
	ball->dx = 0;
	ball->dy = 0;
}

