/************************************************************************** 
 	level.c 
 
  	Project: RockBlast an asteroid game. 
 
	Copyright (c) Febuary 2001 Garett Spencley
 
   	This program is free software; you can redistribute it and/or modify
   	it under the terms of the GNU General Public License as published by
   	the Free Software Foundation; either version 2, or (at your option)
   	any later version.

   	This program is distributed in the hope that it will be useful,
   	but WITHOUT ANY WARRANTY; without even the implied warranty of
   	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   	GNU General Public License for more details.

   	You should have received a copy of the GNU General Public License
   	along with this program; if not, write to the Free Software Foundation,
   	Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 

****************************************************************************/

#include <string.h>
#include <stdlib.h>

#include "level.h"
#include "ball.h"
#include "video.h"
#include "lives.h"
#include "audio.h"
#include "image.h"
#include "number.h"
#include "score.h"
#include "pause.h"
#include "timer.h"

/* Game play stuff */

int ASTEROIDS_LEFT;
bool LEVEL_COMPLETED;
bool PAUSED;

int difficulty_level;
float speed_level;

ball_t ball;
paddle_t paddle;
asteroid_t asteroids[NUM_ASTEROIDS];

object level;
object level_digit1, level_digit2;

void handle_key_events(void);
void draw_level_meter();
void increase_level_meter();
void draw_level_meter();
void erase_level_meter();
char *get_image(int i);

/*
 * next_level()
 */

void next_level()
{
	if(ball.remaining != MAX_LIVES)
		++ball.remaining;

	if(speed_level < MAX_BALL_SPEED)
		speed_level = (float)(speed_level + 0.1);

	++difficulty_level;

	increase_level_meter();
	reset_timer();
}

/*
 * handle_keydown() - handle keyboard events.
 */

void handle_key_events(void)
{
	Uint8 *keys;

	keys = SDL_GetKeyState(NULL);

	if(keys[SDLK_ESCAPE] == SDL_PRESSED) {
		GAME_OVER = true;
		return;
	}

	if(keys[SDLK_LEFT] == SDL_PRESSED) 
		move_paddle(&paddle, PADDLE_LEFT);
	
	if(keys[SDLK_RIGHT] == SDL_PRESSED)
		move_paddle(&paddle, PADDLE_RIGHT);
	
	if(keys[SDLK_SPACE] == SDL_PRESSED) {
		if(!ball.moving) {
			ball.moving = true;
			ball.dx = 1;
			ball.dy = 1;
		}
	}
	
	if(keys[SDLK_a] == SDL_PRESSED)
		add_life(&ball);

	if(keys[SDLK_n] == SDL_PRESSED) {
		ASTEROIDS_LEFT = -1;
		SDL_Delay(100);
		flush_event_queue();
	}
	
	if(keys[SDLK_p] == SDL_PRESSED)
		pause_screen(); //PAUSED = (PAUSED == true ? false : true);

	return;
}

/* 
 * init_level() - set initial values for level
 */

bool init_level()
{
	if(!load_level())
		return false;

	paint_background();

	/* set up objects */

	init_paddle(&paddle);
	draw_paddle(&paddle);

	/* Set up ball */
	ball.obj.x = paddle.obj.x + (paddle.obj.image->w / 2);
	ball.obj.y = (paddle.obj.y - ball.obj.image->h);

	draw_ball(&ball);

	/* lives meter */
	
	if(!init_lives(&ball))
		return false;

	draw_lives(&ball);

	/* Create asteroids */

	if(!init_asteroids())
		return false;

	if(!create_asteroids(asteroids, NUM_ASTEROIDS)) 
		return false;

	draw_asteroids(asteroids, NUM_ASTEROIDS);

	if(!init_timer())
		return false;

	draw_timer(); 

	return true;
}

/* 
 * load_level() - load data for a level
 */

bool load_level()
{
	SDL_Surface *background;

	background = load_image(DATAFILE("background.png"), false);
	if(!background) {
		fprintf(stderr, "%s\n", SDL_GetError());
		return false;
	}
	
	if(!set_background(background)) {
		fprintf(stderr, "%s\n", SDL_GetError());
		return false;
	}

	SDL_FreeSurface(background);

	paddle.obj.image = load_image(DATAFILE("paddle.png"), true);
	if(!paddle.obj.image)
		return false;
	
	ball.obj.image = load_image(DATAFILE("ball.png"), true);
	if(!ball.obj.image)
		return false;
	
	return true;
}

/*
 * free_level() - free memory used by level
 */

void free_level()
{
	SDL_FreeSurface(paddle.obj.image);
	SDL_FreeSurface(ball.obj.image);

	free_asteroids(asteroids, NUM_ASTEROIDS);

	free_timer();
}

/* 
 * run_level() - Runs a level
 */

void run_level()
{
	SDL_Event event;
	int i;
	
	LEVEL_COMPLETED = false;

	while(!LEVEL_COMPLETED && !GAME_OVER && !QUIT) {
			
		wait_frame(FPS);

		/* Erase objects */

		erase_ball(&ball);
		erase_paddle(&paddle);
		erase_asteroids(asteroids, NUM_ASTEROIDS);
		
		erase_level_meter();
		erase_score();
		erase_timer();

		for(i = 0; i < ball.remaining; i++) 
			erase_object(&lives[i]);

		/* Grab events */
	
		handle_key_events();

		/* Snap ball to paddle if the ball isn't moving */

		if(!ball.moving) 
			ball.obj.x = paddle.obj.x + (paddle.obj.image->w / 2);
		else
			move_ball(&ball);
		
		/* Create any special asteroids (like one ups etc.) */

		if(create_special())
			special_asteroid(asteroids, NUM_ASTEROIDS);

		/* update the timer */

		update_timer();
		
		/* Draw Objects */

		draw_paddle(&paddle);
		draw_ball(&ball);

		draw_lives(&ball);

		draw_level_meter();
		draw_score();
		draw_timer();
	
		draw_asteroids(asteroids, NUM_ASTEROIDS);
		
		update_screen();

		if(ASTEROIDS_LEFT == -1)
			LEVEL_COMPLETED = true;

		if(ball.remaining == -1)
			GAME_OVER = true;

		loop_music();

		while(SDL_PollEvent(&event)) {
			if(event.type == SDL_QUIT)
				GAME_OVER = QUIT = true;
		}
	}

	return;
}

/* 
 * init_level_meter() - initializes level meter
 */

bool init_level_meter()
{
	level.image = load_image(DATAFILE("level.png"), true);
	if(!level.image)
		return false;
	
	level.x = 5;
	level.y = screen->h - level.image->h - 2;

	level_digit1.image = num_get_digit(0);
	if(!level_digit1.image) {
		/* this shouldn't happen! */
	}

	level_digit2.image = num_get_digit(0);
	if(!level_digit2.image) {
		/* this shouldn't happen! */
	}

	level_digit1.x = level.x + level.image->w + 10;
	level_digit1.y = screen->h - level_digit1.image->h - 2;

	level_digit2.x = level_digit1.x + level_digit1.image->w + 1;
	level_digit2.y = screen->h - level_digit2.image->h - 2;

	return true;
}

/*
 * draw_level_meter() - draw the level meter
 */

void draw_level_meter()
{
	draw_object(&level);

	draw_object(&level_digit1);
	draw_object(&level_digit2);
}

/*
 * erase_level_meter() - erase the level meter
 */

void erase_level_meter()
{
	erase_object(&level);
	erase_object(&level_digit1);
	erase_object(&level_digit2);
}

/*
 * increase_level_meter() - increse the level meter by one
 */

void increase_level_meter()
{
	int d1, d2;
	
	if(difficulty_level > 99) {
		/* we've capped. god must be playing */
		return;
	}
		
	d1 = difficulty_level / 10;
	d2 = difficulty_level % 10;

	level_digit1.image = num_get_digit(d1);
	level_digit2.image = num_get_digit(d2);
}

/*
 * free_level_meter() - free memory used by level meter
 */

void free_level_meter()
{
	SDL_FreeSurface(level.image);
}

